// uDMX_PcDimm.cpp : Definiert den Einsprungpunkt fr die DLL-Anwendung.
//
// Copyright (C) 2007 Lutz (ilLU[TZ]minator)
//
// Thanx for hints from Christian Noeding himself
//

// uncomment for NEW message-handling
// did not work with all new versions above 4.0
// and is not implemented in version < 4.0
// problem is no response to MASTER FADER 

//#define UDMXPCDIMM_MSG


#include "stdafx.h"

#include "mmsystem.h"
#include "math.h"
#include <oaidl.h>
#include <process.h>

// === CallBacks
// Definitions
typedef void (__stdcall cbSetDLLValues)
                 (int address, int startvalue, 
                  int endvalue, int fadetime) ;
typedef void (__stdcall cbSetDLLValueEvent)
                 (int address, int endvalue) ;
typedef void (__stdcall cbSetDLLNames)
                 (int address, char *channelname) ;
typedef void (__stdcall cbGetDLLValue)
                 (int address) ;
typedef void (__stdcall cbSendMessage)
                 (char MSG, int Data1, int Data2) ;

// Globale Pointers to Functions
cbSetDLLValues      * pSetDLLValues       = NULL ;
cbSetDLLValueEvent  * pSetDLLValueEvent   = NULL ;
cbSetDLLNames       * pSetDLLNames        = NULL ;
cbGetDLLValue       * pGetDLLValue        = NULL ;
cbSendMessage       * pSendMessage        = NULL ;


// === for import of uDMX.dll
typedef BOOL (CALLBACK* PFCHANNELSET)(LONG, LONG) ;
typedef BOOL (CALLBACK* PFCHANNELSSET)(LONG, LONG, LONG *) ;
typedef void (CALLBACK* PFCONFIGURE)() ;
typedef void (CALLBACK* PFINFO)() ;

HMODULE hDLL = NULL ;

PFCHANNELSET  pChannelSet   = NULL ;
PFCHANNELSSET pChannelsSet  = NULL ;
PFCONFIGURE   pConfigure    = NULL ;
PFINFO        pInfo         = NULL ;

bool          bSending  = false ;   // try to avoid blocking

#define       DMX_CHANS   (8 * 512) + 1
#define       TIMER_RES   30

unsigned char dmxStart[DMX_CHANS],
              dmxEnd[DMX_CHANS],
              dmxAkt[DMX_CHANS],
              fading[DMX_CHANS],
              uDMXBuffer[DMX_CHANS];

DWORD         fadeStart[DMX_CHANS],
              fadeEnd[DMX_CHANS] ;

HANDLE        hThreadDMX  = NULL ;
HANDLE        hThreadFade = NULL ;
bool          bRunning    = true ;

// === Thread for managing fading
//     as function DLLSenddata may
//     contain parameters for fading
//     VERY PC-Dimmer specific !
void fadeThread(void *Param)
{
  int   i ;
  DWORD fadeAkt ;

  // Parameter bRunning for ending thread
  while(bRunning)
  {
    // A little break to reduce cpu load
    Sleep(5) ;

    // get actual time
    fadeAkt = GetTickCount() ;

    // check for ALL channels
    for (i = 0; i <= 512; i++)
    {
      // is there some fading ?
      if (fading[i])
      {
        // YES -> fade
        bSending = true ;

        // Has end of fading been reached ?
        if (fadeAkt <= fadeEnd[i])
        {
          // normal fading
          // convert to double because of signed calculation
          uDMXBuffer[i] = (unsigned char)
                          (dmxStart[i] + 
                           (((double)(dmxEnd[i] - dmxStart[i])) * 
                            (fadeAkt - fadeStart[i]) /
                            (fadeEnd[i] - fadeStart[i])) ) ;
        }
        else
        {
          // last step of fading ->
          // set last value AND stop fading
          uDMXBuffer[i] = dmxEnd[i] ;
          fading[i] = false ;
        }

        bSending = false ;
      } // if (fading[i])
    } // for (i
  } // while(bRunning)

  return ;
}

// === Thread for sending to uDMX hardware
void uDMXthread(void *Param)
{
  int i     = 0, 
      iUSBP = 256,
      iMod  = 0 ;

  // Parameter bRunning for ending thread
  while(bRunning)
  {
    // A little break to reduce cpu load
    Sleep(5) ;

    if (pChannelsSet)
    {
      // OUTPUT DATA TO uDMX, if data has changed
      for (i = 1; i <= 512; i = i + iUSBP)
      {
        // Comparing memory blocks
        // reduces transfers on USB
        // for security make a full transfer every 100 times (when idle)
        if ((iMod >= 99) ||
            (memcmp(&uDMXBuffer[i], &dmxAkt[i], iUSBP) != 0))
        {
          // memory has changed (at least 1 byte)
          // Transfer memory block via USB
          if (pChannelsSet(iUSBP, i, (LONG *)&uDMXBuffer[i]))
            // store new data, if transfer was successful
            memcpy(&dmxAkt[i], &uDMXBuffer[i], iUSBP) ;
        } // if (memcmp ...
      } // for ...
    } // if (pChannelsSet)

    // set variable for "idle update" DMX
    iMod++ ;
    if (iMod >= 100)
      iMod = 0 ;
  } // while(bRunning)
}


// === MAIN FUNCTION
BOOL APIENTRY DllMain( HANDLE hModule, 
                       DWORD  ul_reason_for_call, 
                       LPVOID lpReserved)
{
#ifdef UDMXPCDIMM_MSG
  #pragma message (">>>> Kompiliere MIT DLLSendMessage ONLY for PCDimmer > 4.0.x")
  #pragma message (">>>> THIS DOES NOT SUPPORT MASTER FADER IN ALL MY TESTS")
#else
  #pragma message (">>>> Kompiliere OHNE DLLSendMessage")
#endif

  return true ;
}

// === Die folgenden 5 Funktionen koennen vom Plugin aufgerufen werden (s.o.)
//
// procedure CallbackSetDLLValues(address,startvalue,endvalue,fadetime:integer);
//   Funktion wird aufgerufen, wenn Plugins einen neuen Kanalwert an das
//   Hauptprogramm senden sollen.
//
// procedure CallbackSetDLLValueEvent(address,endvalue:integer);
//   Funktion wird aufgerufen, wenn Plugins statt eines neuen Kanalwertes
//   Ereignisse ber die Data-In-Funktion ausfhren sollen.
//
// procedure CallbackSetDLLNames(address:integer; channelname:PChar);
//   Funktion wird aufgerufen, wenn Plugins einen neuen Kanalnamen an das
//   Hauptprogramm senden sollen.
//
// CallbackGetValue(address:integer);
//   Funktion wird aufgerufen, wenn Plugins einen Kanalwert anfordern.
//   Die Funktion ruft (im Hauptprogramm) dann wiederum DLLSenddata()
//   bzw. DLLSendMessage:MSG_ACTUALCHANNELVALUE auf
//
// CallbackSendMessage(MSG:Byte; Data1,Data2: Variant);
//   Funktion kann ausgefhrt werden, um Nachrichten an das Hauptprogramm
//   zu liefern. Derzeit sind folgende Nachrichten registriert (Stand: 26.12.2007):
/*
  MSG_NAME                    =MSG // Data1, Data2
  MSG_RECORDSCENE             =0; // Booleanarray, -
  MSG_AUDIOEFFECTPLAYERRECORD =1; // Boolean, -
  MSG_SYSTEMVOLUME            =2; // Byte, -
  MSG_SYSTEMMUTE              =3; // Boolean, -
  MSG_JUMPTOCHANNEL           =4; // Integer, -
  MSG_GRANDMASTER             =5; // Byte, -
  MSG_FLASHMASTER             =6; // Byte, -
  MSG_ADDLOGFILEENTRY         =7; // String, -
  MSG_EFFECTSPLAY             =8; // Boolean, -
  MSG_SYSTEMSPEED             =9; // Integer, -
  MSG_NEW                     =10; // String (Projektverzeichnis), -
  MSG_OPEN                    =11; // String (Projektverzeichnis), -
  MSG_SAVE                    =12; // String (Projektverzeichnis), -
  MSG_BEATIMPULSE             =13; // Boolean, -
  MSG_ACTUALCHANNELVALUE      =14; // Integer, Integer
  MSG_MIDIIN                  =15; // array[0..2] of byte, -
*/

void __stdcall DLLActivate(cbSetDLLValues * CallbackSetDLLValues,
						               cbSetDLLValueEvent * CallbackSetDLLValueEvent,
                           cbSetDLLNames * CallbackSetDLLNames,
                           cbGetDLLValue * CallbackGetDLLValue,
						               cbSendMessage * CallbackSendMessage)
{
  // Funktionsaufruf bei Pluginaktivierung
  pSetDLLValues     = CallbackSetDLLValues ;
  pSetDLLValueEvent = CallbackSetDLLValueEvent ;
  pSetDLLNames      = CallbackSetDLLNames ;
  pGetDLLValue      = CallbackGetDLLValue ;
  pSendMessage      = CallbackSendMessage ;

  // Load common library uDMX.dll
  hDLL = LoadLibrary("uDMX") ;
  if (hDLL != NULL)
  {
    pChannelSet = (PFCHANNELSET)GetProcAddress(hDLL, "ChannelSet") ;
    pChannelsSet= (PFCHANNELSSET)GetProcAddress(hDLL, "ChannelsSet") ;
    pConfigure  = (PFCONFIGURE) GetProcAddress(hDLL, "Configure") ;    
    pInfo       = (PFINFO)      GetProcAddress(hDLL, "Info") ;
    memset(dmxAkt, 0, DMX_CHANS) ;
    memset(uDMXBuffer, 0, DMX_CHANS) ;
  }

  //SetPriorityClass(GetCurrentProcess(), IDLE_PRIORITY_CLASS) ;
  
  // reduce priority of threads to not block
  hThreadDMX = (HANDLE) _beginthread(uDMXthread, 0, (void *)NULL) ;
  SetThreadPriority(hThreadDMX, /*THREAD_PRIORITY_LOWEST*/ THREAD_PRIORITY_IDLE) ;

  hThreadFade = (HANDLE) _beginthread(fadeThread, 0, (void *)NULL) ;
  SetThreadPriority(hThreadFade, /*THREAD_PRIORITY_LOWEST*/ THREAD_PRIORITY_IDLE) ;


  return ;
}

void __stdcall DLLDeactivate()
{
  // signal threads to end and wait a moment
  bRunning = false ;
  Sleep(100) ;

  // Funktionsaufruf bei Plugindeaktivierung
  if (hDLL)
    FreeLibrary(hDLL) ;

  return ;
}

char * __stdcall DLLIdentify()
{
  return ("Output") ;
}

char * __stdcall DLLGetName()
{
  return ("uDMX Output-Plugin for PC-Dimmer") ;
}

char * __stdcall DLLGetVersion()
{
  return ("V0.2") ;
}

void __stdcall DLLConfigure()
{
  if (!pConfigure)
    MessageBox(NULL, "No configuration available for uDMX Plugin",
               "uDMX-Plugin", MB_OK | MB_ICONSTOP) ;
  else
    pConfigure() ;

  return ;
}

void __stdcall DLLAbout()
{
  // Infodialog
  if (!pInfo)
    MessageBox(NULL, "by ilLU[TZ]minator",
               "uDMX-Plugin", MB_OK | MB_ICONEXCLAMATION);
  else
    pInfo() ;

  return ;
}

// fadetime in milliseconds !
void __stdcall DLLSenddata(int address, 
                           int startvalue, 
                           int endvalue, 
                           int fadetime,
                           char *name)
{
#ifndef UDMXPCDIMM_MSG

  // if already sending end here
  if (bSending)
    return ;

  bSending = true ;

  // initialize fading
  fading[address]     = (fadetime != 0) ;

  if (fading[address])
  {
    // store fading parameters
    fadeStart[address]  = GetTickCount() ;
    fadeEnd[address]    = fadeStart[address] + fadetime ;
  
    dmxStart[address]   = startvalue ;
    dmxEnd[address]     = endvalue ;
  }
  else
  {
    // NO fading -> set output buffer
    // Buffer is send by thread
    uDMXBuffer[address] = endvalue ;
  }

  bSending = false ;

#endif
  return ;
}

bool __stdcall DLLIsSending()
{
  return bSending ;
}

void __stdcall DLLSendMessage(char MSG,
							                VARIANT Adr,
							                VARIANT Value)
{
  // Funktion wird aufgerufen, wenn im PC_DIMMER Nachrichten erzeugt
  // werden. Dies tritt bei Beat-Impulsen, Speicher-/
  // Oeffnen-Anforderungen oder Kanalwertaenderungen des 
  // Dimmerkernels auf. Fuer DMX Geraete sollte auf die Nachricht 
  // MSG = MSG_ACUTALCHANNELVALUE = 14 reagiert und
  // auf Data1=Kanal und Data2=Kanalwert ausgewertet werden.
#ifdef UDMXPCDIMM_MSG
  if (MSG == 14) // MSG_ACTUALCHANNELVALUE=14
  {
    // set output buffer
    uDMXBuffer[Adr.intval] = Value.intVal ;
  }
#endif

  return ;
}

